/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.definition;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.facade.def.impl.ProcessDefinitionImpl;


/**
 * @author Tom Baeyens
 */
public class Node extends CompositeElement {

  private static final long serialVersionUID = 1L;

  protected AbstractActivity behaviourReference;
  protected Node parentNode;

  /**
   * Use {@link ProcessDefinitionImpl#createNode()} or
   * {@link Node#createNode()} instead.
   */
  public Node() {
    super();
  }

  // specialized node containment methods /////////////////////////////////////

  @Override
  public Node addNode(final Node node) {
    node.setParentNode(this);
    super.addNode(node);
    return node;
  }

  @Override
  public Node findNode(final String nodeName) {
    if (nodeName == null) {
      if (this.name == null) {
        return this;
      }
    } else if (nodeName.equals(this.name)) {
      return this;
    }
    return super.findNode(nodeName);
  }

  // behaviour ////////////////////////////////////////////////////////////////

  /**
   * sets the given activity as the behaviour for this node.
   */
  public void setBehaviour(final AbstractActivity activity) {
    this.behaviourReference = activity;
  }


  public AbstractActivity getBehaviour() {
    return this.behaviourReference;
  }

  // various helper methods ///////////////////////////////////////////////////

  static Map<String, Node> getNodesMap(final List<Node> nodes) {
    Map<String, Node> map = null;
    if (nodes != null) {
      map = new HashMap<String, Node>();
      for (final Node node : nodes) {
        if (node.getName() != null) {
          if (!map.containsKey(node.getName())) {
            map.put(node.getName(), node);
          }
        }
      }
    }
    return map;
  }

  @Override
  public String toString() {
    if (this.name != null) {
      return "node(" + this.name + " - " + this.behaviourReference.getName() + ")";
    }
    if (this.dbid != 0) {
      return "node(" + this.dbid + ")";
    }
    return "node(" + System.identityHashCode(this) + ")";
  }

  /**
   * collects the full stack of parent in a list. This node is the first element
   * in the chain. The process definition will be the last element. the chain
   * will never be null.
   */
  public List<CompositeElement> getParentChain() {
    final List<CompositeElement> chain = new ArrayList<CompositeElement>();
    CompositeElement processElement = this;
    while (processElement != null) {
      chain.add(processElement);
      processElement = processElement.getParent();
    }
    return chain;
  }

  // getters and setters //////////////////////////////////////////////////////

  @Override
  public CompositeElement getParent() {
    return this.parentNode != null ? this.parentNode : this.processDefinition;
  }

  @Override
  public String getName() {
    return this.name;
  }

  @Override
  public void setName(final String name) {
    // if there is no processDefinition associated with this node
    if (this.processDefinition == null) {
      // it s just a setter
      this.name = name;

    } else { // otherwise
      // make sure the processDefinition's activitiesMap remains up to date
      if (this.name != null) {
        this.processDefinition.removeNode(this);
      }
      this.name = name;
      if (name != null) {
        this.processDefinition.addNode(this);
      }
    }
  }

  /**
   * retrieve the parent node in the composite node structure. This is different
   * from {@link ObservableElement#getParent()} in that it is restricted to the
   * parent nodes. It doesn't take into account the process definition.
   */
  public Node getParentNode() {
    return this.parentNode;
  }

  public void setParentNode(final Node parentNode) {
    this.parentNode = parentNode;
  }

}
