/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ExecutionState;
import org.ow2.orchestra.util.Misc;

/**
 * the runtime behaviour of the while activity.
 *
 * @author Charles Souillard
 */
public class While extends AbstractBpelActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  private static final Logger LOG = Logger.getLogger(While.class.getName());

  protected Expression condition;

  public While() {
    super();
  }

  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    final Node node = execution.getNode();
    final Node firstNode = node.getNodes().get(0);
    // Execute loop on child execution
    this.executeLoop(execution.createExecution(), firstNode);
  }

  /**
   * Signal method of the NodeBehaviour.
   *
   * @param execution
   *            execution
   * @param signal
   *            signal
   */
  @Override
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null) {
      // this is child execution returning
      final Node node = execution.getNode();
      final Node firstNode = node.getNodes().get(0);
      this.executeLoop(execution, firstNode);
    } else if (signal.equals("finished")) {
      this.afterRunned(execution);
    }

  }

  private void executeLoop(final BpelExecution execution, final Node node) {
    final boolean whileResult = this.condition.getEvaluator().evaluateBoolean(execution);
    if (whileResult) {
      Misc.fastDynamicLog(While.LOG, Level.FINE,"while condition : %s evaluated to true", this.condition);
      execution.execute(node);
    } else {
      // end child execution
      execution.end(ExecutionState.ended);

      final BpelExecution whileExecution = execution.getParent();
      // remove finished execution from parent
      whileExecution.removeExecution(execution);

      // signal must be non-null to make the distinction with child
      // executions
      // in the first line of this method
      whileExecution.signal("finished");
    }
  }

  public Expression getCondition() {
    return this.condition;
  }

  public void setCondition(final Expression condition) {
    this.condition = condition;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.WHILE;
  }

}
