/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.env.descriptor;

import java.lang.management.ManagementFactory;
import java.util.Hashtable;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.ObjectName;

import org.hibernate.SessionFactory;
import org.hibernate.cfg.Configuration;
import org.hibernate.jmx.StatisticsService;
import org.ow2.orchestra.env.Descriptor;
import org.ow2.orchestra.env.WireContext;
import org.ow2.orchestra.env.WireDefinition;
import org.ow2.orchestra.env.WireException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.Listener;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;

/**
 * @author Tom Baeyens
 */
public class HibernateSessionFactoryDescriptor extends AbstractDescriptor {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(HibernateSessionFactoryDescriptor.class.getName());

  private String configurationName;
  private Descriptor configurationDescriptor;

  public Object construct(final WireContext wireContext) {
    Configuration configuration = null;

    if (this.configurationName != null) {
      configuration = (Configuration) wireContext.get(this.configurationName);

    } else if (this.configurationDescriptor != null) {
      configuration = (Configuration) wireContext.create(
          this.configurationDescriptor, false);

    } else {
      configuration = wireContext.get(Configuration.class);
    }

    if (configuration == null) {
      throw new WireException("couldn't find configuration");
    }

    final SessionFactory sessionFactory = configuration.buildSessionFactory();

    wireContext.addListener(new SessionFactoryCloser(sessionFactory));

    final Properties orchestraProperties = (Properties) wireContext.get(OrchestraConstants.ORCHESTRA_PROPERTIES);
    if (orchestraProperties != null) {
      final String jmxObjectName = orchestraProperties.getProperty(OrchestraConstants.JMX_OBJECT_NAME);
      final String domainName = jmxObjectName.substring(0, jmxObjectName.indexOf(":"));

      // Construct mbean name from sessionFactoryName
      String mbeanName = this.getName();
      mbeanName = mbeanName.replace(",", "_");
      mbeanName = mbeanName.replace("=", "_");
      mbeanName = mbeanName.replace(":", "_");
      mbeanName = mbeanName.replace("'", "_");
      mbeanName = mbeanName.replace("*", "_");
      mbeanName = mbeanName.replace("?", "_");
      // Register mbean
      final Hashtable<String, String> tb = new Hashtable<String, String>();
      tb.put("type", "Hibernate");
      tb.put("name", mbeanName);
      final ObjectName on;
      try {
        on = new ObjectName(domainName, tb);
        final StatisticsService stats = new StatisticsService();
        stats.setSessionFactory(sessionFactory);
        ManagementFactory.getPlatformMBeanServer().registerMBean(stats, on);
      } catch (final Exception e) {
        throw new OrchestraRuntimeException("couldn't register mbeans", e);
      }

      wireContext.addListener(new SessionFactoryMBeanCloser(on));
    }
    return sessionFactory;
  }

  public static class SessionFactoryMBeanCloser implements Listener {
    private final ObjectName objectName;

    public SessionFactoryMBeanCloser(final ObjectName on) {
      this.objectName = on;
    }

    public void event(final Object source, final String eventName, final Object info) {
      if (WireContext.EVENT_CLOSE.equals(eventName)) {
        try {
          ManagementFactory.getPlatformMBeanServer().unregisterMBean(this.objectName);
        } catch (final Exception e) {
          throw new OrchestraRuntimeException("Exception while unregistering mbean", e);
        }
      }
    }
  }

  public static class SessionFactoryCloser implements Listener {
    private final SessionFactory sessionFactory;

    public SessionFactoryCloser(final SessionFactory sessionFactory) {
      this.sessionFactory = sessionFactory;
    }

    public void event(final Object source, final String eventName, final Object info) {
      if (WireContext.EVENT_CLOSE.equals(eventName)) {
        Misc.fastDynamicLog(HibernateSessionFactoryDescriptor.LOG, Level.FINE, "closing hibernate session factory");
        this.sessionFactory.close();
      }
    }
  }

  @Override
  public Class< ? > getType(final WireDefinition wireDefinition) {
    return SessionFactory.class;
  }

  public void setConfigurationName(final String configurationName) {
    this.configurationName = configurationName;
  }

  public void setConfigurationDescriptor(final Descriptor configurationDescriptor) {
    this.configurationDescriptor = configurationDescriptor;
  }
}
