/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.If;
import org.ow2.orchestra.definition.element.Else;
import org.ow2.orchestra.definition.element.Elseif;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseIfFullDefinition;
import org.ow2.orchestra.facade.def.full.IfActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.ElseFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.ElseIfFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.IfActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.evaluator.EvaluatorFactory;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard
 */
public class IfBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(IfBinding.class.getName());

  public IfBinding() {
    super("if", ActivityType.IF);
  }

  /**
   * Parses the given dom element to build a Node and a behaviour.
   *
   * @param element dom element to parse
   * @param parse parse
   * @param parser parser
   * @return the build behaviour!
   */
  public Object parse(final Element element, final Parse parse, final Parser parser) {
    if (IfBinding.log.isLoggable(Level.FINE)) {
      IfBinding.log.fine("parsing element = " + element);
    }
    final If ifObject = new If();

    this.setEnclosingScope(ifObject, parse);
    this.parseStandardAttributes(element, ifObject, parse);

    final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    final IfActivityFullDefinition ifActivityFullDefinition = (IfActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();

    this.parseStandardElements(element, ifObject, parse);

    final Node ifNode = this.createActivityNode(parse, parser, ifObject);


    parse.pushObject(ifNode);
    parse.pushObject(ifObject);
    try {
      final Expression expression = this.parseCondition(element, parse);
      ifObject.setCondition(expression);
      final Node nodeImpl = this.parseChildActivity(element, parse, parser);
      ifObject.setActivity(nodeImpl);

      // Definition
      ifActivityFullDefinition.setExpression(expression.getText());
      ifActivityFullDefinition.setExpressionLanguage(expression.getLanguage());

      // elseif elements
      final List<Element> elseifElements = XmlUtil.elements(element, "elseif");
      this.parseElseifs(elseifElements, ifObject, parse, parser, ifActivityFullDefinition);

      // else element
      final Element elseElement = XmlUtil.element(element, "else");
      this.parseElse(elseElement, ifObject, parse, parser, ifActivityFullDefinition);
    } finally {
      parse.popObject();
      parse.popObject();
    }
    return ifNode;
  }

  private void parseElseifs(final List<Element> elseifElements, final If ifObject, final Parse parse, final Parser parser,
    final IfActivityFullDefinition ifActivityFullDefinition) {
    final List<Elseif> elseifs = new ArrayList<Elseif>();
    if (elseifElements != null) {
      for (final Element elseifElement : elseifElements) {
        // Definition
        final ElseIfFullDefinition elseIfFullDefinition = new ElseIfFullDefinitionImpl();

        final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
        defKeeper.pushParent(elseIfFullDefinition);

        final Elseif elseif = this.parseElseif(elseifElement, parse, parser);
        elseifs.add(elseif);

        elseIfFullDefinition.setExpression(elseif.getCondition().getText());
        elseIfFullDefinition.setExpressionLanguage(elseif.getCondition().getLanguage());
        ifActivityFullDefinition.addElseIfFullDefinition(elseIfFullDefinition);
        defKeeper.popParent();
      }
    }
    ifObject.setElseifs(elseifs);
  }

  private Elseif parseElseif(final Element elseifElement, final Parse parse, final Parser parser) {
    final Elseif elseif = new Elseif();
    // condition
    elseif.setCondition(this.parseCondition(elseifElement, parse));
    elseif.setActivity(this.parseChildActivity(elseifElement, parse, parser));
    return elseif;
  }

  private void parseElse(final Element elseElement, final If ifObject, final Parse parse, final Parser parser,
    final IfActivityFullDefinition ifActivityFullDefinition) {
    if (elseElement != null) {

      final ElseFullDefinition elseFullDefinition = new ElseFullDefinitionImpl();
      final DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
      defKeeper.pushParent(elseFullDefinition);

      final Else else1 = new Else();
      else1.setActivity(this.parseChildActivity(elseElement, parse, parser));
      ifObject.setElse(else1);

      ifActivityFullDefinition.setElseFullDefinition(elseFullDefinition);
      defKeeper.popParent();
    }
  }

  private Expression parseCondition(final Element parentElement, final Parse parse) {
    final Element conditionElement = XmlUtil.element(parentElement, "condition");
    if (conditionElement != null) {
      final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
      String expressionLanguage = XmlUtil.attribute(conditionElement, "expressionLanguage");
      if (expressionLanguage == null) {
        expressionLanguage = bpelProcess.getExpressionLanguage();
      }
      final Expression expression = new Expression();
      if (!EvaluatorFactory.isSupportedLanguage(expressionLanguage)) {
        // SA00004
        final String message = StaticAnalysisFault.getMessage("SA00004",
            "  Expression language '" + expressionLanguage + "' not supported.");
        parse.addProblem(message, new StaticAnalysisException("SA00004"), conditionElement);
      }
      expression.setLanguage(expressionLanguage);
      expression.setNamespaces(BpelUtil.getAllNameSpaces(conditionElement));
      expression.setText(conditionElement.getTextContent());
      return expression;
    }
    return null;
  }

  @Override
  public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new IfActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }

}
