package org.ow2.orchestra.persistence.db.hibernate;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.hibernate.Cache;
import org.hibernate.FlushMode;
import org.hibernate.HibernateException;
import org.hibernate.LockMode;
import org.hibernate.LockOptions;
import org.hibernate.Query;
import org.hibernate.ScrollableResults;
import org.hibernate.Session;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.BpelProcess.InstanceCreationLock;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.deployment.ActiveProcessHolder;
import org.ow2.orchestra.deployment.ProcessDeploymentCounter;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.db.RuntimeDbSession;
import org.ow2.orchestra.runtime.BpelInstance;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.ResultIterator;
import org.ow2.orchestra.services.job.ExecuteInvokeJob;
import org.ow2.orchestra.services.job.Job;
import org.ow2.orchestra.services.jobexecutor.JobExecutor;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReceivingElementUtil;
import org.w3c.dom.Document;

public class RuntimeDbSessionImpl extends
    HibernateDbSession implements RuntimeDbSession {

  private static final Logger LOG = Logger.getLogger(RuntimeDbSessionImpl.class.getName());
  private static final String PROCESS_QUERY_CACHE = "processQueryCache";

  public RuntimeDbSessionImpl(final Session session) {
    this.setSession(session);
    session.setFlushMode(FlushMode.COMMIT);
  }

  public ActiveProcessHolder findActiveBpelProcess(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getActiveProcessHolder");
    query.setString("qname", processQName.toString());
    query.setCacheable(true);
    query.setCacheRegion(RuntimeDbSessionImpl.PROCESS_QUERY_CACHE);
    return (ActiveProcessHolder) query.uniqueResult();
  }

  public List<ProcessDefinitionUUID> findBpelProcesses(final QName processQName, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessesUUIDByQNameForLock");
      query.setString("qname", processQName.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final List<ProcessDefinitionUUID> res = query.list();
      return res;
    }
    return this.findBpelProcesses(processQName);
  }

  private List<ProcessDefinitionUUID> findBpelProcesses(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("findBpelProcessesUUIDByQName");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    query.setCacheable(true);
    return query.list();
  }

  public BpelProcess findBpelProcess(final ProcessDefinitionUUID pduuid, final boolean lock) {
    if (lock) {
      final Query query = this.getSession().getNamedQuery("findBpelProcessForLock");
      query.setString("uuid", pduuid.toString());
      query.setLockOptions(new LockOptions(LockMode.PESSIMISTIC_WRITE));
      // don't use cache for this request or the lock
      query.setCacheable(false);
      final InstanceCreationLock res = (InstanceCreationLock) query.uniqueResult();
      if (res == null) {
        return null;
      }
    }
    return this.getBpelProcess(pduuid);
  }

  public BpelProcess getBpelProcess(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelProcessByUUID");
    query.setString("uuid", pduuid.toString());
    query.setCacheable(true);
    final BpelProcess processDefinition = (BpelProcess) query.uniqueResult();
    return processDefinition;
  }

  public List<ProcessDefinitionUUID> getBpelProcessesUUID() {
    final Query query = this.getSession().getNamedQuery("getBpelProcesses");
    query.setCacheable(true);
    return query.list();
  }

  private static class ResultIteratorImpl<T> implements ResultIterator<T> {

    private final ScrollableResults results;
    private final Session session;
    private T current = null;

    public ResultIteratorImpl(final ScrollableResults results, final Session session) {
      this.results = results;
      this.session = session;
    }
    public T next() {
      return next(true);
    }

    public T next(final boolean evict) {
      if (this.results == null) {
        return null;
      }
      if (this.current != null && evict) {
        this.session.evict(this.current);
      }
      if (this.results.next()) {
        try {
          this.current = (T) this.results.get(0);
        } catch (final HibernateException e) {
          this.current = null;
        }
        if (this.current == null) {
          Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO, "WaitingExecution not found, skipping.");
          return this.next();
        }
        return this.current;
      } else {
        this.results.close();
      }
      return null;
    }
  }

  public ResultIterator<WaitingExecution> getWaitingExecutions(
      final ReceivingElement receivingElement,
      final Map<QName, Document> csProperties) {
    final List<Set<QName>> propertiesValues = ReceivingElementUtil.getCorrelationPropertyNamesPermutations(receivingElement);

    Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO,
        "getWaitingExecutions:  %s possible correlations found.", propertiesValues.size());

    for (final Set<QName> propQNames : propertiesValues) {
      Misc.fastDynamicLog(RuntimeDbSessionImpl.LOG, Level.INFO,
          "getWaitingExecutions:  trying to match %s correlation properties.", propQNames.size());
      final Map<QName, Document> csProps = new HashMap<QName, Document>(csProperties);
      csProps.keySet().retainAll(propQNames);

      final int correlationKey = ReceivingElementUtil.getCorrelationPropertiesKey(csProps);

      final Query query = this.getSession().getNamedQuery("getWaitingExecutionsByCorrelationKey");
      query.setParameter("activity", receivingElement.getInboundMessageElement().getUUID().toString());
      query.setInteger("correlationKey", correlationKey);

      final ScrollableResults waitingExecutions = query.scroll();
      if (waitingExecutions.first()) {
        waitingExecutions.beforeFirst();
        return new ResultIteratorImpl<WaitingExecution>(waitingExecutions, this.getSession());
      } else {
        waitingExecutions.close();
      }
    }
    return new ResultIteratorImpl<WaitingExecution>(null, this.getSession());
  }

  public WaitingExecution getFirstWaitingExecution(final InboundMessageElement inboundMessageElement) {
    final Query query = this.getSession().getNamedQuery("getWaitingExecutions");
    query.setParameter("activity", inboundMessageElement.getUUID().toString());
    query.setMaxResults(1);
    return (WaitingExecution) query.uniqueResult();
  }

  public ResultIterator<PendingMessage> getPendingMessages(final OperationKey operationKey, final Integer correlationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessagesByCorrelationKey");
    query.setString("operationKey", operationKey.toString());
    query.setInteger("correlationKey", correlationKey);
    final ScrollableResults pendingMessages = query.scroll();
    if (pendingMessages.first()) {
      pendingMessages.beforeFirst();
      return new ResultIteratorImpl<PendingMessage>(pendingMessages, this.getSession());
    } else {
      pendingMessages.close();
      return new ResultIteratorImpl<PendingMessage>(null, this.getSession());
    }
  }

  public ResultIterator<PendingMessage> getAllPendingMessages(final OperationKey operationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessages");
    query.setString("operationKey", operationKey.toString());
    final ScrollableResults pendingMessages = query.scroll();
    if (pendingMessages.first()) {
      pendingMessages.beforeFirst();
      return new ResultIteratorImpl<PendingMessage>(pendingMessages, this.getSession());
    } else {
      pendingMessages.close();
      return new ResultIteratorImpl<PendingMessage>(null, this.getSession());
    }
  }

  public PendingMessage getFirstPendingMessage(final OperationKey operationKey) {
    final Query query = this.getSession().getNamedQuery("getPendingMessages");
    query.setString("operationKey", operationKey.toString());
    query.setMaxResults(1);
    return (PendingMessage) query.uniqueResult();
  }


  public Collection<PendingMessage> getPendingMessages() {
    final Query query = this.getSession().getNamedQuery("getAllPendingMessages");
    return query.list();
  }

  public List<BpelInstance> getBpelInstances() {
    final Query query = this.getSession().getNamedQuery("getBpelInstances");
    return query.list();
  }

  public BpelInstance getBpelInstance(final ProcessInstanceUUID processInstanceUUID) {
    final Query query = this.getSession().getNamedQuery("getBpelInstance");
    query.setString("processInstanceUUID", processInstanceUUID.toString());
    final BpelInstance bpelExecution = (BpelInstance) query.uniqueResult();
    return bpelExecution;
  }

  public List<BpelInstance> getBpelInstances(final QName processQName) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcess");
    query.setString("namespace", processQName.getNamespaceURI());
    query.setString("name", processQName.getLocalPart());
    return query.list();
  }

  public List<BpelInstance> getBpelInstances(final ProcessDefinitionUUID pduuid) {
    final Query query = this.getSession().getNamedQuery("getBpelInstancesOfProcessByUUID");
    query.setString("uuid", pduuid.toString());
    return query.list();
  }

  public ProcessDeploymentCounter getProcessDeploymentCounter() {
    return (ProcessDeploymentCounter) this.getSession().createCriteria(ProcessDeploymentCounter.class).uniqueResult();
  }

  public List<Job> findJobsWithException() {
    // query definition can be found at the bottom resource
    // org/ow2/orchestra/pvm/hibernate.job.hbm.xml
    final Query query = this.session.getNamedQuery("findJobsWithException");
    return query.list();
  }

  public Collection<ExecuteInvokeJob> findLockedExecuteInvokeJobs() {
    final Query query = this.session.getNamedQuery("findLockedExecuteInvokeJobs");
    query.setDate("lockDate", ExecuteInvokeJob.LOCK_TIMEOUT_DATE);
    query.setString("lockOwner", Environment.getFromCurrent(JobExecutor.class).getName());
    return query.list();
  }

  public void clearActiveProcessCache(final QName processQName) {
    final Cache cache = this.getSession().getSessionFactory().getCache();
    if (cache != null) {
      final ActiveProcessHolder activeProcessHolder = this.findActiveBpelProcess(processQName);
      if (activeProcessHolder != null) {
        cache.evictEntity(ActiveProcessHolder.class, activeProcessHolder.getDbid());
      }
      cache.evictQueryRegion(RuntimeDbSessionImpl.PROCESS_QUERY_CACHE);

    }
  }
}
