/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.data.runtime.PartnerLinkUpdateData;
import org.ow2.orchestra.facade.runtime.PartnerLinkUpdateType;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * runtime representation of one partner for a {@link PartnerLinkRole}.
 *
 * @author Tom Baeyens, Charles SOuillard
 */
public class PartnerLinkRuntime {

  protected long dbid;
  protected int dbversion;

  protected ScopeRuntime scopeRuntime;
  protected String name;

  /**
   * The value of the endpointReference associated to the partnerRole of the
   * partnerLink. It is a mechanism that allows you to select one of the
   * available services in the WSDL, or even define new services at runtime.
   */
  protected Document partnerRoleEndPointReference;

  protected PartnerLinkRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param partnerLinkDefinition
   *            partnerLinkDefinition of this partnerLinkRuntime.
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to.
   */
  public PartnerLinkRuntime(final PartnerLink partnerLinkDefinition,
      final BpelExecution execution, final ScopeRuntime scopeRuntime) {
    this.scopeRuntime = scopeRuntime;
    this.name = partnerLinkDefinition.getName();
    if (partnerLinkDefinition.getMyRolePortTypeQName() != null) {
      this.initializeMyRole(execution, partnerLinkDefinition);
    }
    if (partnerLinkDefinition.getPartnerRolePortTypeQName() != null) {
      this.initializePartnerRole(execution, partnerLinkDefinition);
    }
  }

  /**
   * Initialize partnerRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  public void initializePartnerRole(final BpelExecution execution, final PartnerLink partnerLinkDefinition) {
    if (partnerLinkDefinition.getDefaultPartnerRoleEndPointReference() != null) {
      final AbstractActivity activity = execution.getNode().getBehaviour();
      EnvTool.getRecorder().recordPartnerLinkUpdated(
        new PartnerLinkUpdateData(
            execution.getActivityInstanceUUID(),
            activity.createRuntimeData(execution),
            partnerLinkDefinition.getName(),
            PartnerLinkUpdateType.PARTNER_ROLE_INITIALIZED,
            partnerLinkDefinition.getDefaultPartnerRoleEndPointReference().getDocumentElement()));
    }
  }

  /**
   * Initialize myRole endpoint reference to the defaultPort value.
   *
   * @param execution
   *            execution to which this partnerLinkRuntime belongs to
   */
  public void initializeMyRole(final BpelExecution execution, final PartnerLink partnerLinkDefinition) {
    if (partnerLinkDefinition.getDefaultMyRoleEndPointReference() != null) {
      final AbstractActivity activity = execution.getNode().getBehaviour();
      EnvTool.getRecorder().recordPartnerLinkUpdated(
        new PartnerLinkUpdateData(
            execution.getActivityInstanceUUID(),
            activity.createRuntimeData(execution),
            partnerLinkDefinition.getName(),
            PartnerLinkUpdateType.MY_ROLE_INITIALIZED,
            partnerLinkDefinition.getDefaultMyRoleEndPointReference().getDocumentElement()));
    }
  }

  public Element getMyRoleEndPointReference(final BpelExecution execution) {
    final Document endpoint = this.getPartnerLinkDefinition(execution).getDefaultMyRoleEndPointReference();
    if (endpoint != null) {
      return endpoint.getDocumentElement();
    }
    return null;
  }

  public Element getPartnerRoleEndPointReference(final BpelExecution execution) {
    if (this.partnerRoleEndPointReference != null) {
      return this.partnerRoleEndPointReference.getDocumentElement();
    }
    final Document endpoint = this.getPartnerLinkDefinition(execution).getDefaultPartnerRoleEndPointReference();
    if (endpoint != null) {
      return endpoint.getDocumentElement();
    }
    return null;
  }

  public void setPartnerRoleEndPointReference(
      final Element partnerRoleEPR,
      final ActivityInstanceData activityInstanceData) {
    final Document doc = XmlUtil.getNewDocument();
    doc.appendChild(doc.importNode(partnerRoleEPR, true));
    this.partnerRoleEndPointReference = doc;
    EnvTool.getRecorder().recordPartnerLinkUpdated(
      new PartnerLinkUpdateData(
          this.scopeRuntime.getActivityInstanceUUID(),
          activityInstanceData,
          this.getName(),
          PartnerLinkUpdateType.PARTNER_ROLE_UPDATED,
          this.partnerRoleEndPointReference.getDocumentElement()));
  }

  public String getName() {
    return this.name;
  }

  public PartnerLink getPartnerLinkDefinition(final BpelExecution execution) {
    final Scope scope = (Scope) execution.getProcessDefinition().getNode(this.scopeRuntime.getScopeNodeUUID()).getBehaviour();
    return scope.getPartnerLinks().get(this.name);
  }

  public long getDbid() {
    return this.dbid;
  }

}
