/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import java.util.Set;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.runtime.ScopeStateUpdateData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.runtime.ScopeState;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.util.EnvTool;

/**
 * @author Charles Souillard
 */
public class ScopeRuntime {

  /**
   * @author Guillaume Porcher
   *
   */
  private static final class ScopeRuntimeEndTimeComparator implements Comparator<ScopeRuntime>, Serializable {

    public int compare(final ScopeRuntime o1, final ScopeRuntime o2) {
      return Long.valueOf(o1.endedTime).compareTo(Long.valueOf(o2.endedTime));
    }
  }

  protected long dbid;
  protected int dbversion;

  /** scope definition of this scopeRuntime. */
  protected String scopeNodeUUID = null;

  /** children scope runtime */
  protected List<ScopeRuntime> endedChildren = new ArrayList<ScopeRuntime>();
  protected ScopeRuntime parent;
  protected long endedTime;

  /** all runtime variables of this scope. */
  protected Map<String, VariableRuntime> variables = null;

  /** all partnerLinkRuntime of this scope. */
  protected Map<String, PartnerLinkRuntime> partnerLinks = null;

  /** all correlationSetRuntime of this scope. */
  protected Map<String, CorrelationSetRuntime> correlationSets = null;

  /** all open IMAs of this scope */
  protected Set<Replier> repliers = new HashSet<Replier>();

  /** UUID of Scope Activity instance */
  protected ActivityInstanceUUID activityInstanceUUID;

  /** state of this scope runtime. */
  protected ScopeState state = ScopeState.INITIALIZATION_ERROR;

  /** last exception thrown in this scope runtime. */
  protected BpelFaultException lastException = null;

  /** Compensate execution to signal when compensation handler is finished **/
  protected BpelExecution compensateExecution;

  protected ScopeRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param scopeDefinition scopeDefinition
   * @param execution execution that this scope belongs to
   */
  public ScopeRuntime(final Scope scopeDefinition, final BpelExecution execution) {
    this.activityInstanceUUID = execution.getActivityInstanceUUID();
    this.scopeNodeUUID = scopeDefinition.getScopeNode().getName();
    // scope children
    this.createVariables(scopeDefinition, execution);
    this.createPartnerLinks(scopeDefinition, execution);
    this.createCorrelationSets(scopeDefinition, execution);

    this.state = ScopeState.INITIALIZED;
  }

  void addEndedChild(final ScopeRuntime scopeRuntime) {
    scopeRuntime.setParent(this);
    scopeRuntime.setEndedTime(System.currentTimeMillis());
    this.getEndedChildren().add(scopeRuntime);
  }

  void setEndedTime(final long endedTime) {
    this.endedTime = endedTime;
  }

  /**
   * @param parent the parent to set
   */
  public void setParent(final ScopeRuntime parent) {
    this.parent = parent;
  }


  /**
   * @return the endedChildren
   */
  public List<ScopeRuntime> getEndedChildren() {
    return this.endedChildren;
  }

  /**
   * for each variable defined in scopeDefinition, create a new runtime
   * variable.
   *
   * @param execution execution that this scope belongs to
   */
  private void createVariables(final Scope scopeDefinition, final BpelExecution execution) {
    this.variables = new HashMap<String, VariableRuntime>();
    final Map<String, Variable> scopeVariables = scopeDefinition.getVariables();
    if (scopeVariables != null) {
      for (final Variable variable : scopeVariables.values()) {
        // Object variableValue = getVariableInitialValue(variable, execution);
        this.variables.put(variable.getName(), new VariableRuntime(variable, execution, this));
      }
    }
  }

  public Collection<String> getVariableNames() {
    if (this.variables == null) {
      return Collections.emptySet();
    } else {
      return this.variables.keySet();
    }
  }

  public VariableRuntime createVariable(final BpelExecution execution, final Variable variable) {
    final VariableRuntime variableRuntime = new VariableRuntime(variable, execution, this);
    this.variables.put(variable.getName(), variableRuntime);
    return variableRuntime;
  }

  public void removeVariable(final BpelExecution execution, final Variable variable) {
    this.variables.remove(variable.getName());
  }

  /**
   * for each partnerLink defined in scopeDefinition, create a new runtime
   * partnerLink.
   *
   * @param execution execution that this scope belongs to
   */
  private void createPartnerLinks(final Scope scopeDefinition, final BpelExecution execution) {
    this.partnerLinks = new HashMap<String, PartnerLinkRuntime>();
    final Map<String, PartnerLink> scopePartnerLinks = scopeDefinition.getPartnerLinks();
    if (scopePartnerLinks != null) {
      for (final PartnerLink partnerLink : scopePartnerLinks.values()) {
        final PartnerLinkRuntime partnerLinkRuntime = new PartnerLinkRuntime(partnerLink, execution, this);
        this.partnerLinks.put(partnerLink.getName(), partnerLinkRuntime);
      }
    }
  }

  /**
   * for each correlationSet defined in scopeDefinition, create a new runtime
   * correlationSet.
   *
   * @param execution execution that this scope belongs to
   */
  private void createCorrelationSets(final Scope scopeDefinition, final BpelExecution execution) {
    this.correlationSets = new HashMap<String, CorrelationSetRuntime>();
    final Map<String, CorrelationSet> correlationSetsMap = scopeDefinition.getCorrelationSets();
    if (correlationSetsMap != null) {
      for (final CorrelationSet correlationSet : correlationSetsMap.values()) {
        final CorrelationSetRuntime correlationSetRuntime = new CorrelationSetRuntime(correlationSet, this);
        this.correlationSets.put(correlationSet.getName(), correlationSetRuntime);
      }
    }
  }

//  public Scope getScopeDefinition() {
//    return this.scopeDefinition;
//  }

  /**
   * Returns the partnerLinkRuntime having the given name.
   *
   * @param partnerLinkName partnerLinkName
   * @return the partnerLinkRuntime having the given name.
   */
  public PartnerLinkRuntime getPartnerLink(final String partnerLinkName) {
    return this.getPartnerLinks().get(partnerLinkName);
  }

  /**
   * Returns the CorrelationSetRuntime having the given name.
   *
   * @param csName csName
   * @return the CorrelationSetRuntime having the given name.
   */
  public CorrelationSetRuntime getCorrelationSet(final String csName) {
    return this.getCorrelationSets().get(csName);
  }

  /**
   * @return the correlationSets
   */
  public Map<String, CorrelationSetRuntime> getCorrelationSets() {
    return this.correlationSets;
  }


  /**
   * @return the partnerLinks
   */
  public Map<String, PartnerLinkRuntime> getPartnerLinks() {
    return this.partnerLinks;
  }

  /**
   * Returns the VariableRuntime having the given name.
   *
   * @param variableName variableName
   * @return the VariableRuntime having the given name.
   */
  public VariableRuntime getVariable(final String variableName) {
    return this.getVariables().get(variableName);
  }

  /**
   * @return the variables
   */
  public Map<String, VariableRuntime> getVariables() {
    return this.variables;
  }

  /**
   * return true if this scope contains a variableRuntime with the given
   * variableName.
   *
   * @param variableName variableName
   * @return true if this scope contains a variableRuntime with the given
   *         variableName.
   */
  public boolean containsVariable(final String variableName) {
    return this.getVariables().containsKey(variableName);
  }

  @Override
  public int hashCode() {
    return this.activityInstanceUUID.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof ScopeRuntime)) {
      return false;
    }
    final ScopeRuntime other = (ScopeRuntime) obj;
    return other.getActivityInstanceUUID().equals(this.getActivityInstanceUUID());
  }

  public List<ScopeRuntime> getEndedChildrenOrderedByEndTime() {
    final List<ScopeRuntime> result = new ArrayList<ScopeRuntime>(this.endedChildren);
    Collections.sort(result, new ScopeRuntimeEndTimeComparator());
    return result;
  }

  public ScopeState getState() {
    return this.state;
  }

  public boolean isCompenseable() {
    return ScopeState.COMPLETED_SUCCESSFUL.equals(this.state);
  }

  public void setState(final ScopeState newState, final BpelExecution execution) {
    this.state = newState;
    final AbstractActivity activity = execution.getNode().getBehaviour();
    EnvTool.getRecorder().recordScopeStateChanged(
        new ScopeStateUpdateData(this.activityInstanceUUID, activity.createRuntimeData(execution), newState));
  }

  public BpelFaultException getLastException() {
    return this.lastException;
  }

  public void setLastException(final BpelFaultException lastException) {
    this.lastException = lastException;
  }

  public ActivityInstanceUUID getActivityInstanceUUID() {
    return this.activityInstanceUUID;
  }

  public BpelExecution getCompensateExecution() {
    return this.compensateExecution;
  }

  public void setCompensateExecution(final BpelExecution compensateExecution) {
    this.compensateExecution = compensateExecution;
  }


  /**
   * @return the scopeNodeUUID
   */
  public String getScopeNodeUUID() {
    return this.scopeNodeUUID;
  }

  /** Get the list of open IMAs */
  public Set<Replier> getRepliers() {
    return this.repliers;
  }

  public void addReplier(final Replier replier) {
    this.getRepliers().add(replier);
  }
}
