/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services;

import java.util.UUID;
import java.util.concurrent.TimeoutException;

import javax.transaction.Synchronization;
import javax.xml.namespace.QName;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.var.MessageVariable;

public class MessageCarrierImpl implements MessageCarrier {

  private final Object lock = new Object();

  private MessageVariable message;

  private QName faultQName;

  private boolean hasMessage = false;
  private boolean isEnded = false;
  private boolean isInstanceEnded = false;

  private final String uuid;

  public MessageCarrierImpl() {
    this.uuid = UUID.randomUUID().toString();
  }

  /**
   * Wait until the message has arrived.
   * @return message
   */
  public MessageVariable getMessage() {
    try {
      return this.getMessage(0);
    } catch (final TimeoutException e) {
      Misc.unreachableStatement("No timeout should occur");
      return null;
    }
  }

  /**
   * Wait until the message has arrived or timeout occurs.
   * @param timeout timeout in millisecond.
   * If timeout is set to 0 then this method has the same behaviour as {@link #getMessage()}
   * @return message
   * @throws TimeoutException thrown if no message has arrived when timeout occurs.
   */
  public MessageVariable getMessage(final long timeout) throws TimeoutException {
    synchronized (this.lock) {
      if (!this.isEnded) {
        try {
          this.lock.wait(timeout);
          if (!this.isEnded) {
            throw new TimeoutException("Timeout occured while waiting for message");
          }
        } catch (final InterruptedException e) {
          throw new OrchestraRuntimeException(e);
        }
      }
      if (!this.hasMessage) {
        if (isInstanceEnded) {
          throw new OrchestraRuntimeException(MessageCarrierImpl.NO_REPLY_INSTANCE_FAILED_MESSAGE);
        } else {
          throw new OrchestraRuntimeException(MessageCarrierImpl.NO_REPLY_NO_PROCESS_MESSAGE);
        }
      }
      return this.message;
    }
  }

  /**
   * @return the faultQName
   */
  public QName getFaultQName() {
    return this.faultQName;
  }

  private void replyFaultMessage(final QName faultName, final MessageVariable messageToSend) {
    synchronized (this.lock) {
      this.faultQName = faultName;
      this.message = messageToSend;
      this.hasMessage = true;
      this.isEnded = true;
      this.lock.notifyAll();
    }
  }

  private void replyMessage(final MessageVariable messageToSend) {
    synchronized (this.lock) {
      this.message = messageToSend;
      this.hasMessage = true;
      this.isEnded = true;
      this.lock.notifyAll();
    }
  }

  private void noReplyException(final boolean isInstanceEnded) {
    synchronized (this.lock) {
      this.hasMessage = false;
      this.isEnded = true;
      this.isInstanceEnded = isInstanceEnded;
      this.lock.notifyAll();
    }
  }

  public void setMessage(final MessageVariable messageToSend) {
    final Transaction tx = Environment.getCurrent().get(Transaction.class);
    if (tx != null) {
      final Synchronization synchro = new Synchronization() {

        private final MessageVariable msg = messageToSend.duplicate();

        public void beforeCompletion() {
          if (MessageCarrierImpl.this.isEnded) {
            throw new OrchestraRuntimeException("Message already associated");
          }
        }
        public void afterCompletion(final int status) {
          if (status == javax.transaction.Status.STATUS_COMMITTED) {
            MessageCarrierImpl.this.replyMessage(this.msg);
          }
        }
      };
      tx.registerSynchronization(synchro);
    } else {
      this.replyMessage(messageToSend);
    }
  }

  public void setFaultMessage(final QName faultName, final MessageVariable messageToSend) {
    final Transaction tx = Environment.getCurrent().get(Transaction.class);
    if (tx != null) {
      final Synchronization synchro = new Synchronization() {

        private final MessageVariable msg = messageToSend.duplicate();

        public void beforeCompletion() {
          if (MessageCarrierImpl.this.isEnded) {
            throw new OrchestraRuntimeException("Message already associated");
          }
        }
        public void afterCompletion(final int status) {
          if (status == javax.transaction.Status.STATUS_COMMITTED) {
            MessageCarrierImpl.this.replyFaultMessage(faultName, this.msg);
          }
        }
      };
      tx.registerSynchronization(synchro);
    } else {
      this.replyFaultMessage(faultName, messageToSend);
    }
  }

  public void throwMissingReply(final boolean isInstanceExited) {
    final Transaction tx = Environment.getCurrent().get(Transaction.class);
    if (tx != null) {
      final Synchronization synchro = new Synchronization() {

        public void beforeCompletion() {
          if (MessageCarrierImpl.this.isEnded) {
            throw new OrchestraRuntimeException("Message already associated");
          }
        }
        public void afterCompletion(final int status) {
          if (status == javax.transaction.Status.STATUS_COMMITTED) {
            MessageCarrierImpl.this.noReplyException(isInstanceExited);
          }
        }
      };
      tx.registerSynchronization(synchro);
    } else {
      this.noReplyException(isInstanceExited);
    }
  }

  public boolean hasMessage() {
    synchronized (this.lock) {
      return this.hasMessage;
    }
  }

  public boolean hasFault() {
    synchronized (this.lock) {
      return this.faultQName != null;
    }
  }

  public String getUuid() {
    return this.uuid;
  }

}
