/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import javax.xml.namespace.QName;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.deployment.ActiveProcessHolder;
import org.ow2.orchestra.deployment.ProcessDeploymentCounter;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.data.runtime.PendingMessageData;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.persistence.db.RuntimeDbSession;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.BpelInstance;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.WaitingExecution;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.ResultIterator;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.services.job.ExecuteInvokeJob;
import org.ow2.orchestra.services.job.Job;
import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.util.EnvTool;
import org.w3c.dom.Document;

public class DbRepository implements Repository {

  /*
   * PROCESS REPOSITORY
   */
  public long getNextProcessCount() {
    ProcessDeploymentCounter processDeploymentCounter = this.getBpelPersistenceService().getProcessDeploymentCounter();
    if (processDeploymentCounter == null) {
      processDeploymentCounter = new ProcessDeploymentCounter();
      this.getBpelPersistenceService().save(processDeploymentCounter);
    }
    return processDeploymentCounter.getAndIncrementProcessCounter();
  }

  public BpelProcess getActiveProcess(final QName processQName) {
    final ActiveProcessHolder activeProcessHolder = this.getBpelPersistenceService().findActiveBpelProcess(processQName);
    if (activeProcessHolder != null) {
      final ProcessDefinitionUUID activeProcessUUID = activeProcessHolder.getProcessUUID();
      if (activeProcessUUID != null) {
        return this.getProcess(activeProcessUUID);
      }
    }
    return null;
  }

  public BpelProcess getProcess(final ProcessDefinitionUUID processUUID) {
    final PublisherImpl publisherImpl = Environment.getFromCurrent(PublisherImpl.class, false);
    if (publisherImpl != null) {
      final BpelProcess process = publisherImpl.getBpelProcess(processUUID);
      if (process != null) {
        return process;
      }
    }
    return this.getBpelPersistenceService().getBpelProcess(processUUID);
  }

  public List<BpelProcess> getProcesses(final QName processQName) {
    final List<ProcessDefinitionUUID> uuids = this.getBpelPersistenceService().findBpelProcesses(processQName, false);
    return this.getBpelProcesses(uuids);
  }

  public List<BpelProcess> getProcesses() {
    return this.getBpelProcesses(this.getBpelPersistenceService().getBpelProcessesUUID());
  }

  public BpelProcess removeProcess(final QName processQName) {
    final BpelProcess process = this.getActiveProcess(processQName);
    if (process != null) {
      this.getBpelPersistenceService().delete(process);
      final PublisherImpl publisherImpl = Environment.getFromCurrent(PublisherImpl.class, false);
      if (publisherImpl != null) {
        publisherImpl.removeBpelProcess(process);
      }
    }
    return process;
  }

  public BpelProcess removeProcess(final ProcessDefinitionUUID pduuid) {
    final BpelProcess process = this.getBpelPersistenceService().getBpelProcess(pduuid);
    if (process != null) {
      this.getBpelPersistenceService().delete(process);
      final List<BpelProcess> otherVersions = this.getProcesses(process.getQName());
      otherVersions.remove(process);
      if (otherVersions.isEmpty()) {
        final ActiveProcessHolder activeProcessHolder =
          this.getBpelPersistenceService().findActiveBpelProcess(process.getQName());
        this.getBpelPersistenceService().delete(activeProcessHolder);
      }
    }
    return process;
  }

  public void storeProcess(final BpelProcess process) {
    this.getBpelPersistenceService().save(process);
    ActiveProcessHolder activeProcessHolder = this.getBpelPersistenceService().findActiveBpelProcess(process.getQName());
    if (activeProcessHolder == null) {
      // Add new process holder if none defined.
      activeProcessHolder = new ActiveProcessHolder(process.getQName());
      this.getBpelPersistenceService().save(activeProcessHolder);
    }
    activeProcessHolder.setProcessUUID(process.getUUID());
  }

  public List<BpelProcess> lockProcesses(final QName processQName) {
    final List<ProcessDefinitionUUID> uuids = this.getBpelPersistenceService().findBpelProcesses(processQName, true);
    return this.getBpelProcesses(uuids);
  }

  private List<BpelProcess> getBpelProcesses(final List<ProcessDefinitionUUID> uuids) {
    if (uuids.isEmpty()) {
      return Collections.emptyList();
    }
    final List<BpelProcess> result = new ArrayList<BpelProcess>();
    for (final ProcessDefinitionUUID processUUID : uuids) {
      result.add(this.getProcess(processUUID));
    }
    return result;
  }
  /*
   * INSTANCE REPOSITORY
   */
  public BpelInstance getInstance(final ProcessInstanceUUID processInstanceUUID) {
    return this.getBpelPersistenceService().getBpelInstance(processInstanceUUID);
  }

  public List<BpelInstance> getInstances() {
    return this.getBpelPersistenceService().getBpelInstances();
  }

  public List<BpelInstance> getInstances(final QName processQName) {
    return this.getBpelPersistenceService().getBpelInstances(processQName);
  }

  public List<BpelInstance> getInstances(final ProcessDefinitionUUID pduuid) {
    return this.getBpelPersistenceService().getBpelInstances(pduuid);
  }

  public void removeInstance(final BpelInstance instance) {
    this.removeInstance(instance, false);
  }

  public void storeExecution(final BpelExecution child) {
    this.getBpelPersistenceService().save(child);
  }
  public void removeExecution(final BpelExecution child) {
    this.getBpelPersistenceService().delete(child);
  }

  public void storeScopeRuntime(final ScopeRuntime scopeRuntime) {
    this.getBpelPersistenceService().save(scopeRuntime);
  }
  public void removeScopeRuntime(final ScopeRuntime scopeRuntime) {
    this.getBpelPersistenceService().delete(scopeRuntime);
  }

  public void storeWaitingExecution(final WaitingExecution waitingExecution) {
    this.getBpelPersistenceService().save(waitingExecution);
  }
  public void removeWaitingExecution(final WaitingExecution waitingExecution) {
    this.getBpelPersistenceService().delete(waitingExecution);
  }

  public void removeInstance(final BpelInstance instance, final boolean force) {
    if (force) {
      if (instance != null
          && (instance.getProcessInstanceState().equals(ActivityState.RUNNING)
              || instance.getProcessInstanceState().equals(ActivityState.SUSPENDED))
      ) {
        instance.exit();
      }
    }
    if (instance != null) {
      this.getBpelPersistenceService().delete(instance);
    }
  }

  public void storeInstance(final BpelInstance instance) {
    this.getBpelPersistenceService().save(instance.getProcessInstance());
  }

  public BpelExecution getUpdatedExecution(final long executionId) {
    return this.getBpelPersistenceService().get(BpelExecution.class, Long.valueOf(executionId));
  }

  public List<Job> findJobsWithException() {
    return this.getBpelPersistenceService().findJobsWithException();
  }

  public JobImpl< ? > getJob(final long jobId) {
    return this.getBpelPersistenceService().get(JobImpl.class, Long.valueOf(jobId));
  }

  public Collection<ExecuteInvokeJob> getLockedExecuteInvokeJobs() {
    return this.getBpelPersistenceService().findLockedExecuteInvokeJobs();
  }

  public ResultIterator<WaitingExecution> getWaitingExecutions(
      final ReceivingElement receivingElement,
      final Map<QName, Document> csProperties) {
    return this.getBpelPersistenceService().getWaitingExecutions(receivingElement, csProperties);
  }

  public WaitingExecution getFirstWaitingExecution(final InboundMessageElement inboundMessageElement) {
    return this.getBpelPersistenceService().getFirstWaitingExecution(inboundMessageElement);
  }

  public void storePendingMessage(final PendingMessage pendingMessage) {
    this.getBpelPersistenceService().save(pendingMessage);
  }

  public void removePendingMessage(final PendingMessage pendingMessage) {
    this.getBpelPersistenceService().delete(pendingMessage);
  }

  public ResultIterator<PendingMessage> getPendingMessages(final OperationKey operationKey,
                                                           final Integer correlationKey) {
    return this.getBpelPersistenceService().getPendingMessages(operationKey, correlationKey);
  }

  public ResultIterator<PendingMessage> getAllPendingMessages(final OperationKey operationKey) {
    return this.getBpelPersistenceService().getAllPendingMessages(operationKey);
  }

  public PendingMessage getFirstPendingMessage(final OperationKey operationKey) {
    return this.getBpelPersistenceService().getFirstPendingMessage(operationKey);
  }

  public Collection<PendingMessage> getPendingMessages() {
    return this.getBpelPersistenceService().getPendingMessages();
  }

  public PendingMessage getPendingMessage(final PendingMessageData pendingMessageData) {
    return this.getBpelPersistenceService().get(PendingMessage.class, pendingMessageData.getMessageId());
  }

  /*
  * COMMON METHODS
  */
  private RuntimeDbSession getBpelPersistenceService() {
    return Environment.getCurrent().get(RuntimeDbSession.class);
  }

  public void activateProcess(final BpelProcess bpelProcess) {
    this.changeProcessState(bpelProcess, false);
  }
  public void retireProcess(final BpelProcess bpelProcess) {
    this.changeProcessState(bpelProcess, true);
  }
  private void changeProcessState(final BpelProcess bpelProcess, final boolean retired) {
    final ActiveProcessHolder activeProcessHolder =
      this.getBpelPersistenceService().findActiveBpelProcess(bpelProcess.getQName());
    if (bpelProcess.getUUID().equals(activeProcessHolder.getProcessUUID())) {
      // process was active
      if (retired) {
        // remove from activeProcessHolder
        activeProcessHolder.setProcessUUID(null);
        // process state changed
        EnvTool.getRecorder().recordProcessStateChanged(bpelProcess.getProcessDefinitionData());
      }
    } else {
      // process was retired
      if (!retired) {
        // set as active in activeProcessHolder (desactivates other versions)
        activeProcessHolder.setProcessUUID(bpelProcess.getUUID());
        // process state changed
        EnvTool.getRecorder().recordProcessStateChanged(bpelProcess.getProcessDefinitionData());
      }
    }
  }

  public void refreshActiveProcess(final ProcessDefinitionUUID pduuid) {
    final BpelProcess process = this.getProcess(pduuid);
    if (process != null) {
      this.getBpelPersistenceService().clearActiveProcessCache(process.getQName());
    }
  }
}
