/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.job;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.Set;

import javax.wsdl.Definition;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.InvokeExecutor;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.impl.PublisherImpl;
import org.ow2.orchestra.var.MessageVariable;
import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * @author Guillaume Porcher
 *
 */
public class ExecuteInvokeJob extends MessageImpl<Void> {

  private static final long serialVersionUID = 7011019390393185222L;

  public static class ExecuteInvokeJobData {

    private MessageVariable requestMessage;
    private Document partnerEndPointReference;
    private OperationKey operationKey;

    public ExecuteInvokeJobData() {
    }
    public MessageVariable getRequestMessage() {
      return this.requestMessage;
    }
    public void setRequestMessage(final MessageVariable requestMessage) {
      this.requestMessage = requestMessage;
    }
    public Document getPartnerEndPointReference() {
      return this.partnerEndPointReference;
    }
    public void setPartnerEndPointReference(final Document partnerEndPointReference) {
      this.partnerEndPointReference = partnerEndPointReference;
    }
    public OperationKey getOperationKey() {
      return this.operationKey;
    }
    public void setOperationKey(final OperationKey operationKey) {
      this.operationKey = operationKey;
    }
  }

  public static final Date LOCK_TIMEOUT_DATE = new GregorianCalendar(4242, Calendar.JANUARY, 1).getTime();

  private final ExecuteInvokeJobData data = new ExecuteInvokeJobData();

  /* needed for hibernate */
  protected ExecuteInvokeJob() { }

  public ExecuteInvokeJob(
      final MessageVariable requestMessage,
      final Element partnerEndPointReference,
      final OperationKey operationKey) {
    this.data.setRequestMessage(requestMessage != null ? requestMessage.duplicate() : null);
    if (partnerEndPointReference != null) {
      this.data.setPartnerEndPointReference(partnerEndPointReference.getOwnerDocument());
    }
    this.data.setOperationKey(operationKey);
  }

  public Void execute(final Environment environment) {

    // set lock to max to ensure the job is not re-executed
    this.setLockExpirationTime(ExecuteInvokeJob.LOCK_TIMEOUT_DATE);

    final BpelExecution bpelExecution = (BpelExecution) this.getExecution();

    final Set<Definition> wsdlDefinitions;
    final PublisherImpl publisherImpl = environment.get(PublisherImpl.class);
    if (publisherImpl != null) {
      wsdlDefinitions = publisherImpl.getWsdlDefinitions(bpelExecution.getProcessDefinition().getUUID());
    } else {
      wsdlDefinitions = new HashSet<Definition>(bpelExecution.getProcessDefinition().getWsdlInfos().getWsdlDefinitions());
    }


    // launch partner service invocation
    environment.get(InvokeExecutor.class).executeInvoke(
        environment,
        bpelExecution,
        this.data.getOperationKey(),
        this.data.getPartnerEndPointReference().getDocumentElement(),
        this.data.getRequestMessage(),
        wsdlDefinitions);

    // don't delete this job,
    // this will be deleted by InvokeExecutor when service invocation is ended.
    // This allows to recover on restart.
    return null;
  }


}
