/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.jobexecutor;

import java.util.Collection;
import java.util.Date;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.hibernate.StaleStateException;
import org.hibernate.exception.GenericJDBCException;
import org.hibernate.exception.LockAcquisitionException;
import org.ow2.orchestra.services.commands.Command;
import org.ow2.orchestra.services.commands.CommandService;
import org.ow2.orchestra.services.commands.RetryInterceptor;
import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.util.PrefixThreadFactory;

/**
 * Common part of dispatcher thread implementations.
 * This thread is responsible for acquiring jobs in the jobImpl that need to be
 * executed and to give them to the jobExecutor thread pool.
 * There is only one dispatcher thread per JobExecutor.
 *
 * @author Tom Baeyens
 * @author Guillaume Porcher
 *
 */
public abstract class AbstractDispatcherThread extends Thread {

  private static final Logger LOG = Logger.getLogger(AbstractDispatcherThread.class.getName());
  private final ThreadPoolExecutor addWaitingJobExecutor = new ThreadPoolExecutor(
      1,
      1,
      60L,
      TimeUnit.SECONDS,
      new LinkedBlockingQueue<Runnable>(),
      new PrefixThreadFactory("addWaitingJobExecutor"));;

  /**
   * @author Guillaume Porcher
   *
   */
  private final class AddJobTask implements Runnable {

    private final JobImpl< ? > job;
    /**
     * @param job
     */
    private AddJobTask(final JobImpl< ? > job) {
      this.job = job;
    }
    public void run() {
      AbstractDispatcherThread.this.getJobExecutor()
        .getDispatcherThread().putAcquiredJobOnQueue(this.job);
    }
  }

  private boolean isActive = true;
  private boolean checkForNewJobs;
  private final Object semaphore = new Object();

  /**
   *
   */
  public AbstractDispatcherThread() {
    super();
  }

  /**
   * @param target
   */
  public AbstractDispatcherThread(final Runnable target) {
    super(target);
  }

  /**
   * @param name
   */
  public AbstractDispatcherThread(final String name) {
    super(name);
  }

  /**
   * @param group
   * @param target
   */
  public AbstractDispatcherThread(final ThreadGroup group, final Runnable target) {
    super(group, target);
  }

  protected abstract void putAcquiredJobOnQueue(JobImpl<?> acquiredJob);

  /**
   * @param group
   * @param name
   */
  public AbstractDispatcherThread(final ThreadGroup group, final String name) {
    super(group, name);
  }

  /**
   * @param target
   * @param name
   */
  public AbstractDispatcherThread(final Runnable target, final String name) {
    super(target, name);
  }

  /**
   * @param group
   * @param target
   * @param name
   */
  public AbstractDispatcherThread(final ThreadGroup group, final Runnable target, final String name) {
    super(group, target, name);
  }

  /**
   * @param group
   * @param target
   * @param name
   * @param stackSize
   */
  public AbstractDispatcherThread(final ThreadGroup group, final Runnable target, final String name, final long stackSize) {
    super(group, target, name, stackSize);
  }

  @Override
  public void run() {
    AbstractDispatcherThread.LOG.log(Level.INFO, "starting...");
    long delay = 50;
    int delayFactor = 2;
    final CommandService commandService = this.getJobExecutor().getCommandExecutor();
    if (commandService instanceof RetryInterceptor) {
      delay = ((RetryInterceptor) commandService).getDelay();
      delayFactor = ((RetryInterceptor) commandService).getDelayFactor();
    }
    long currentDelay = delay;
    try {
      while (this.isActive) {
        try {
          // checkForNewJobs is set to true in jobWasAdded() below
          this.checkForNewJobs = false;
          // try to acquire jobs
          final Collection<JobImpl<?>> acquiredJobs = this.acquireJobs();

          // no exception so resetting the currentIdleInterval
          currentDelay = delay;
          if ((acquiredJobs != null) && (!acquiredJobs.isEmpty())) {
            for (final JobImpl<?> job : acquiredJobs) {
              this.putAcquiredJobOnQueue(job);
              if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
                AbstractDispatcherThread.LOG.log(Level.FINEST, "added job " + job + " to the queue");
              }
            }

          } else if (this.isActive) {
            final long waitPeriod = this.getWaitPeriod();
            if (waitPeriod > 0) {
              synchronized (this.semaphore) {
                if (!this.checkForNewJobs) {
                  if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
                    AbstractDispatcherThread.LOG.log(Level.FINEST,
                        this.getName() + " will wait for max " + waitPeriod + " ms on " + this.getJobExecutor());
                  }
                  this.semaphore.wait(waitPeriod);
                  if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
                    AbstractDispatcherThread.LOG.log(Level.FINEST, this.getName() + " woke up");
                  }
                } else {
                  if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
                    AbstractDispatcherThread.LOG.log(Level.FINEST, "skipped wait because new message arrived");
                  }
                }
              }
            }
          }

        } catch (final InterruptedException e) {
          if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
            AbstractDispatcherThread.LOG.log(Level.FINEST,
                (this.isActive ? "active" : "inactivated")
                + " jobImpl dispatcher thread '" + this.getName() + "' got interrupted");
          }
        } catch (final Exception e) {
          final Level level;
          if (e instanceof StaleStateException
              || e instanceof LockAcquisitionException
              || e instanceof GenericJDBCException) {
            level = Level.FINEST;
          } else {
            level = Level.SEVERE;
          }
          if (AbstractDispatcherThread.LOG.isLoggable(level)) {
            AbstractDispatcherThread.LOG.log(level, "exception in job executor dispatcher thread."
                + " Waiting " + currentDelay + " milliseconds",e);
          }
          try {
            synchronized (this.semaphore) {
              this.semaphore.wait(currentDelay);
            }
          } catch (final InterruptedException e2) {
            if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
              AbstractDispatcherThread.LOG.log(Level.FINEST,
                  "delay after exception got interrupted", e2);
            }
          }
          // after an exception, the current idle interval is doubled to prevent
          // continuous exception generation when e.g. the db is unreachable
          currentDelay = currentDelay * delayFactor;
        }
      }
    } finally {
      AbstractDispatcherThread.LOG.log(Level.INFO, this.getName() + " leaves cyberspace");
    }
  }

  protected Collection<JobImpl<?>> acquireJobs() {
    final CommandService commandService = this.getJobExecutor().getCommandExecutor();
    final Command<Collection<JobImpl<?>>> acquireJobsCommand = this.getJobExecutor().getAcquireJobsCommand();
    return commandService.execute(acquireJobsCommand);
  }

  protected Date getNextDueDate() {
    final CommandService commandService = this.getJobExecutor().getCommandExecutor();
    final Command<Date> getNextDueDate = this.getJobExecutor().getNextDueDateCommand();
    return commandService.execute(getNextDueDate);
  }

  protected long getWaitPeriod() {
    long interval = this.getJobExecutor().getIdleMillis();
    final Date nextDueDate = this.getNextDueDate();
    if (nextDueDate != null) {
      final long currentTimeMillis = System.currentTimeMillis();
      final long nextDueDateTime = nextDueDate.getTime();
      if (nextDueDateTime < currentTimeMillis + interval) {
        interval = nextDueDateTime - currentTimeMillis;
      }
    }
    if (interval < 0) {
      interval = 0;
    }
    return interval;
  }

  public void deactivate() {
    this.deactivate(false);
  }

  public void deactivate(final boolean join) {
    if (this.isActive) {
      AbstractDispatcherThread.LOG.log(Level.FINE, "deactivating " + this.getName());
      this.isActive = false;
      this.addWaitingJobExecutor.shutdown();
      this.interrupt();
      if (join) {
        while (this.isAlive()) {
          try {
            AbstractDispatcherThread.LOG.log(Level.FINEST, "joining " + this.getName());
            this.join();
            this.addWaitingJobExecutor.awaitTermination(180, TimeUnit.SECONDS);
          } catch (final InterruptedException e) {
            AbstractDispatcherThread.LOG.log(Level.FINEST, "joining " + this.getName() + " got interrupted");
          }
        }
      }
    } else {
      AbstractDispatcherThread.LOG.log(Level.FINEST, "ignoring deactivate: " + this.getName() + " is not active");
    }
  }

  public void jobWasAdded() {
    if (AbstractDispatcherThread.LOG.isLoggable(Level.FINEST)) {
      AbstractDispatcherThread.LOG.log(Level.FINEST, "notifying jobImpl executor dispatcher thread of new jobImpl");
    }
    synchronized (this.semaphore) {
      this.checkForNewJobs = true;
      this.semaphore.notify();
    }
  }

  public void addWaitingJob(final JobImpl<?> job) {
    this.addWaitingJobExecutor.execute(new AddJobTask(job));
  }


  public ThreadPoolExecutor getAddWaitingJobExecutor() {
    return this.addWaitingJobExecutor;
  }

  public boolean isActive() {
    return this.isActive;
  }

  public abstract AbstractJobExecutor getJobExecutor();
}