/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.jobexecutor;

import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Status;
import javax.transaction.Synchronization;

import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.util.Misc;

/**
 * listener that can be registered as a listener to the transaction to notify
 * the jobImpl executor of added jobs so that the dispatcher thread can wake up.
 *
 * @author Tom Baeyens
 */
public class JobAddedNotification implements Synchronization {

  private static final Logger LOG = Logger.getLogger(JobAddedNotification.class.getName());

  private final JobExecutor jobExecutor;
  private final JobImpl<?> job;

  private final boolean addToQueue;

  public JobAddedNotification(final JobExecutor jobExecutor, final JobImpl<?> job) {
    this.jobExecutor = jobExecutor;
    this.job = job;
    this.addToQueue = this.job.getDueDate().before(new Date())
        && (this.jobExecutor instanceof AbstractJobExecutor)
        && this.jobExecutor.isActive()
        && !((AbstractJobExecutor) this.jobExecutor).isLimitSameInstanceJobs();
  }

  public void afterCompletion(final int status) {
    if (status == Status.STATUS_COMMITTED) {
      Misc.fastDynamicLog(JobAddedNotification.LOG, Level.FINEST, "notifying jobImpl executor of added message");
      if (this.job.getLockExpirationTime() != null) {
        ((AbstractJobExecutor) this.jobExecutor).getDispatcherThread().addWaitingJob(this.job);
      } else {
        this.jobExecutor.jobWasAdded();
      }
    }
  }

  public void beforeCompletion() {
    if (this.addToQueue) {
      if (this.jobExecutor instanceof AbstractJobExecutor) {
        final AbstractJobExecutor pvmJobExecutor = (AbstractJobExecutor) this.jobExecutor;
        if (pvmJobExecutor.getDispatcherThread().getAddWaitingJobExecutor().getQueue().size() >= pvmJobExecutor.getAddWaitingJobQueueSize()) {
          return;
        }
      }
      this.job.setLockExpirationTime(
          new Date(System.currentTimeMillis() + ((AbstractJobExecutor) this.jobExecutor).getLockMillis()));
      this.job.setLockOwner(this.jobExecutor.getName());
    }
  }

  @Override
  public String toString() {
    return "job-added-notification";
  }
}
