/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.wsdl.Part;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Charles Souillard
 */
public final class BpelUtil {

  /**
   * Default private constructor to prevent this class to be instanciated.
   */
  private BpelUtil() { }

  /**
   * Return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   * @param element dom element to analyse
   * @return true if the given element has an attribute named createInstance and if
   * its value is 'yes'.
   */
  public static boolean isCreateInstance(final Element element) {
    final String createInstance = XmlUtil.attribute(element, "createInstance");
    if (createInstance != null && createInstance.equals("yes")) {
      return true;
    }
    return false;
  }
  /**
   * Get all namespaces defined in the given node and all of its parents.
   * @param n node from where namespaces are taken
   * @return all namespaces defined in the given node and all of its parents.
   */
  public static Map<String, String> getAllNameSpaces(final org.w3c.dom.Node n) {
    final HashMap<String, String> result = new HashMap<String, String>();
    if (n == null) {
      return result;
    }
    //we look for the NameSpace in the parent Node
    if (n.getParentNode() != null) {
      final Map<String, String> resultParent = BpelUtil.getAllNameSpaces(n.getParentNode());
      for (final Entry<String, String> ns : resultParent.entrySet()) {
        if (!result.containsKey(ns.getKey())) {
          result.put(ns.getKey(), ns.getValue());
        }
      }

    }
    final NamedNodeMap attributes = n.getAttributes();
    if (attributes != null) {
      for (int i = 0; i < attributes.getLength(); i++) {
        if (attributes.item(i).getNodeName().contains("xmlns:")) {
          final String nodeName = attributes.item(i).getNodeName();
          final String prefix = nodeName.split(":")[1];
          result.put(prefix, attributes.item(i).getTextContent());
        }
      }
    }
    result.put(XmlConstants.BPEL_FUNC_PREFIX, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE);
    return result;
  }


  /**
   * Look for an instance of a class in a list of ExtensibilityElement.
   * @param listOfExtensibilityElement the list of ExtensibilityElement.
   * @param classToFind : the class we llok for
   * @return If found, return the element found, else, return null.
   */
  public static ExtensibilityElement findExtensibilityElement(
      final List<ExtensibilityElement> listOfExtensibilityElement,
      final Class<? extends ExtensibilityElement> classToFind) {
    for (final ExtensibilityElement extElem : listOfExtensibilityElement) {
      if (classToFind.isInstance(extElem)) {
        return extElem;
      }
    }
    return null;
  }
  /**
   * return true if the given ns is a xmlSchema.
   * @param ns ns to analyse
   * @return true if the given ns is a xmlSchema.
   */
  public static boolean isXmlSchemaNs(final String ns) {
    return ns != null && (ns.equals(XmlConstants.XMLNS_XMLSCHEMA) || ns.equals(XmlConstants.XMLNS_XMLSCHEMAXSD));
  }
  /**
   * Return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   * @param address address to use to build the serviceRef document
   * @param serviceQN serviceQN to use to build the serviceRef document
   * @return a dom document conformant with ServiceRef xsd based on the given address and serviceQName.
   */
  public static Document getServiceRefDocument(final SOAPAddress address, final QName serviceQN) {
    final Document document = XmlUtil.getNewDocument();

    final Element serviceRef =
      document.createElementNS(XmlConstants.XMLNS_BPEL_2_0_SERVICEREF, XmlConstants.SERVICE_REF_TAG_NAME);
    document.appendChild(serviceRef);

    final Element epr =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE, XmlConstants.EPR_TAG_NAME);
    serviceRef.appendChild(epr);

    final Element addressElement =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE,XmlConstants.ADDRESS_TAG_NAME);
    addressElement.setTextContent(address.getLocationURI());
    epr.appendChild(addressElement);

    final Element serviceNameElement =
      document.createElementNS(XmlConstants.ADDRESSING_2004_REF_NAMESPACE,XmlConstants.SERVICE_NAME_TAG_NAME);
    serviceNameElement.setAttribute("xmlns:eprServicePrefix", serviceQN.getNamespaceURI());
    serviceNameElement.setTextContent("eprServicePrefix:" + serviceQN.getLocalPart());
    epr.appendChild(serviceNameElement);

    return document;
  }


  public static Element getPartElement(final Element father, final Part part) {
    final String partName = part.getName();
    final List<Element> childrenElements = XmlUtil.elements(father);
    for (final Element child : childrenElements) {
      if (child.getLocalName().equals(partName)) {
        return child;
      }
    }
    return null;
  }

  public static boolean booleanValue(final String text, final boolean defaultValue) {
    final Boolean result = BpelUtil.booleanValue(text);
    if (result == null) {
      return defaultValue;
    }
    return result.booleanValue();
  }

  public static Boolean booleanValue(final String text) {
    if (text == null) {
      return null;
    }

    // if we have to check for value true
    if (("true".equalsIgnoreCase(text))
        || ("enabled".equalsIgnoreCase(text))
        || ("on".equalsIgnoreCase(text))
        || ("yes".equalsIgnoreCase(text))
    ) {
      return Boolean.TRUE;
    }
    if (("false".equalsIgnoreCase(text))
        || ("disabled".equalsIgnoreCase(text))
        || ("off".equalsIgnoreCase(text))
        || ("no".equalsIgnoreCase(text))
    ) {
      return Boolean.FALSE;
    }

    return null;
  }
}


