package org.ow2.orchestra.util;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.Enumeration;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.env.descriptor.ArgDescriptor;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;

public abstract class ReflectUtil {
  private static final Logger LOG = Logger.getLogger(ReflectUtil.class.getName());

  private ReflectUtil() { }

  static ClassLoader resolveClassLoader(ClassLoader classLoader) {
    // 1) if the user provided a classloader through the API, use that one
    if (classLoader != null) {
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "using provided classloader");
      return classLoader;
    }

    // 2) if the user provided a classloader through the environment, use that
    // one
    final Environment environment = Environment.getCurrent();
    if (environment != null) {
      classLoader = environment.getClassLoader();
      if (classLoader != null) {
        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "using environment classloader");
        return classLoader;
      }
    }

    // 3) otherwise, use the current thread's context classloader
    Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "using context classloader");
    return Thread.currentThread().getContextClassLoader();
  }

  public static Class< ? > loadClass(ClassLoader classLoader, final String className) {
    try {
      classLoader = ReflectUtil.resolveClassLoader(classLoader);
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "loading class %s", className);
      return classLoader.loadClass(className);
    } catch (final NoClassDefFoundError e) {
      throw new OrchestraRuntimeException("couldn't define class " + className, e);
    } catch (final ClassNotFoundException e) {
      throw new OrchestraRuntimeException("couldn't load class " + className, e);
    }
  }

  public static InputStream getResourceAsStream(ClassLoader classLoader,
      final String resource) {
    classLoader = ReflectUtil.resolveClassLoader(classLoader);
    Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "getting resource as stream %s", resource);
    return classLoader.getResourceAsStream(resource);
  }

  public static Enumeration<URL> getResources(ClassLoader classLoader,
      final String resource) {
    classLoader = ReflectUtil.resolveClassLoader(classLoader);
    try {
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "getting resources %s", resource);
      return classLoader.getResources(resource);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't get resources " + resource, e);
    }
  }

  public static URL getResource(ClassLoader classLoader, final String resource) {
    classLoader = ReflectUtil.resolveClassLoader(classLoader);
    try {
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "getting resource %s", resource);
      return classLoader.getResource(resource);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't get resource " + resource, e);
    }
  }

  public static Object instantiate(ClassLoader classLoader, final String className) {
    Object newObject;
    try {
      classLoader = ReflectUtil.resolveClassLoader(classLoader);
      final Class< ? > clazz = ReflectUtil.loadClass(classLoader, className);
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "instantiating %s", className);
      newObject = clazz.newInstance();
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't instantiate " + className, e);
    }
    return newObject;
  }

  public static Class< ? >[] loadClasses(ClassLoader classLoader,
      final List<String> constructorArgTypeNames) {
    if (constructorArgTypeNames == null) {
      return null;
    }
    final Class< ? >[] classes = new Class[constructorArgTypeNames.size()];
    for (int i = 0; i < constructorArgTypeNames.size(); i++) {
      classLoader = ReflectUtil.resolveClassLoader(classLoader);
      classes[i] = ReflectUtil.loadClass(classLoader, constructorArgTypeNames.get(i));
    }
    return classes;
  }

  public static <T> Constructor<T> getConstructor(final Class<T> clazz,
      final Class< ? >[] parameterTypes) {
    Constructor<T> constructor = null;
    try {
      constructor = clazz.getDeclaredConstructor(parameterTypes);

      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "found constructor %s(%s)", clazz.getName(), ArrayUtil.toString(parameterTypes));

    } catch (final SecurityException e) {
      throw new OrchestraRuntimeException("wasn't allowed to get constructor '"
          + clazz.getName() + "(" + ReflectUtil.getParameterTypesText(parameterTypes)
          + ")'", e);
    } catch (final NoSuchMethodException e) {
      throw new OrchestraRuntimeException("couldn't find constructor '" + clazz.getName()
          + "(" + ReflectUtil.getParameterTypesText(parameterTypes) + ")'", e);
    }

    return constructor;
  }

  public static Field getField(final Class< ? > clazz, final String fieldName) {
    return ReflectUtil.getField(clazz, fieldName, clazz);
  }

  private static Field getField(final Class< ? > clazz, final String fieldName,
      final Class< ? > original) {
    Field field = null;

    try {
      field = clazz.getDeclaredField(fieldName);
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "found field %s in %s", fieldName, clazz.getName());
    } catch (final SecurityException e) {
      throw new OrchestraRuntimeException("wasn't allowed to get field '" + clazz.getName()
          + "." + fieldName + "'", e);
    } catch (final NoSuchFieldException e) {
      if (clazz.getSuperclass() != null) {
        return ReflectUtil.getField(clazz.getSuperclass(), fieldName, original);
      } else {
        throw new OrchestraRuntimeException("couldn't find field '" + original.getName()
            + "." + fieldName + "'", e);
      }
    }

    return field;
  }

  public static Method getMethod(final Class< ? > clazz, final String methodName,
      final Class< ? >[] parameterTypes) {
    return ReflectUtil.getMethod(clazz, methodName, parameterTypes, clazz);
  }

  private static Method getMethod(final Class< ? > clazz, final String methodName,
      final Class< ? >[] parameterTypes, final Class< ? > original) {
    Method method = null;

    try {
      method = clazz.getDeclaredMethod(methodName, parameterTypes);

      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
          "found method %s.%s(%s)", clazz.getName(), methodName, ArrayUtil.toString(parameterTypes));

    } catch (final SecurityException e) {
      throw new OrchestraRuntimeException("wasn't allowed to get method '" + clazz.getName()
          + "." + methodName + "(" + ReflectUtil.getParameterTypesText(parameterTypes)
          + ")'", e);
    } catch (final NoSuchMethodException e) {
      if (clazz.getSuperclass() != null) {
        return ReflectUtil.getMethod(clazz.getSuperclass(), methodName, parameterTypes,
            original);
      } else {
        throw new OrchestraRuntimeException("couldn't find method '" + original.getName()
            + "." + methodName + "(" + ReflectUtil.getParameterTypesText(parameterTypes)
            + ")'", e);
      }
    }

    return method;
  }

  private static String getParameterTypesText(final Class< ? >[] parameterTypes) {
    if (parameterTypes == null) {
      return "";
    }
    final StringBuffer parametersTypeText = new StringBuffer();
    for (int i = 0; i < parameterTypes.length; i++) {
      final Class< ? > parameterType = parameterTypes[i];
      parametersTypeText.append(parameterType.getName());
      if (i != parameterTypes.length - 1) {
        parametersTypeText.append(", ");
      }
    }
    return parametersTypeText.toString();
  }

  public static <T> T newInstance(final Class<T> clazz) {
    return ReflectUtil.newInstance(clazz, null, null);
  }

  public static <T> T newInstance(final Constructor<T> constructor) {
    return ReflectUtil.newInstance(null, constructor, null);
  }

  public static <T> T newInstance(final Constructor<T> constructor, final Object[] args) {
    return ReflectUtil.newInstance(null, constructor, args);
  }

  private static <T> T newInstance(final Class<T> clazz, Constructor<T> constructor,
      final Object[] args) {
    if ((clazz == null) && (constructor == null)) {
      throw new IllegalArgumentException(
          "can't create new instance without clazz or constructor");
    }

    String className = null;
    try {
      if (constructor == null) {
        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "getting default constructor");
        constructor = clazz.getConstructor((Class[]) null);
      }
      className = constructor.getDeclaringClass().getName();
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
          "creating new instance for class '%s' with args %s", className, ArrayUtil.toString(args));
      if (!constructor.isAccessible()) {
        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "making constructor accessible");
        constructor.setAccessible(true);
      }
      return constructor.newInstance(args);

    } catch (final Exception t) {
      throw new OrchestraRuntimeException("couldn't construct new '" + className
          + "' with args " + ArrayUtil.toString(args), t);
    }
  }

  public static Object get(final Field field, final Object object) {
    if (field == null) {
      throw new NullPointerException("field is null");
    }
    try {
      final Object value = field.get(object);
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
          "got value '%s' from field '%s'", value, field.getName());
      return value;
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't get '" + field.getName() + "'", e);
    }
  }

  public static void set(final Field field, final Object object, final Object value) {
    if (field == null) {
      throw new NullPointerException("field is null");
    }
    try {
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
          "setting field '%s' to value '%s'", field.getName(), value);
      if (!field.isAccessible()) {
        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "making field accessible");
        field.setAccessible(true);
      }
      field.set(object, value);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't set '" + field.getName() + "' to '"
          + value + "'", e);
    }
  }

  public static Object invoke(final Method method, final Object target, final Object[] args) {
    if (method == null) {
      throw new OrchestraRuntimeException("method is null");
    }
    try {
      Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
          "invoking '%s' on '%s' with %s", method.getName(), target, ArrayUtil.toString(args));
      if (!method.isAccessible()) {
        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST, "making method accessible");
        method.setAccessible(true);
      }
      return method.invoke(target, args);
    } catch (final InvocationTargetException e) {
      final Throwable targetException = e.getTargetException();
      throw new OrchestraRuntimeException("couldn't invoke '" + method.getName() + "' with "
          + ArrayUtil.toString(args) + " on " + target + ": "
          + targetException.getMessage(), targetException);
    } catch (final Exception e) {
      throw new OrchestraRuntimeException("couldn't invoke '" + method.getName() + "' with "
          + ArrayUtil.toString(args) + " on " + target + ": " + e.getMessage(),
          e);
    }
  }

  public static Method findMethod(final Class< ? > clazz, final String methodName,
      final List<ArgDescriptor> argDescriptors, final Object[] args) {
    Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
        "searching for method %s in %s", methodName, clazz.getName());
    final Method[] candidates = clazz.getDeclaredMethods();
    for (final Method candidate : candidates) {
      if ((candidate.getName().equals(methodName))
          && (ReflectUtil.isArgumentMatch(candidate.getParameterTypes(), argDescriptors,
              args))) {

        Misc.fastDynamicLog(ReflectUtil.LOG, Level.FINEST,
            "found matching method %s.%s", clazz.getName(), methodName);
        return candidate;
      }
    }
    if (clazz.getSuperclass() != null) {
      return ReflectUtil.findMethod(clazz.getSuperclass(), methodName, argDescriptors, args);
    }
    return null;
  }

  public static Constructor< ? > findConstructor(final Class< ? > clazz,
      final List<ArgDescriptor> argDescriptors, final Object[] args) {
    final Constructor< ? >[] constructors = clazz.getDeclaredConstructors();
    for (final Constructor< ? > constructor : constructors) {
      if (ReflectUtil.isArgumentMatch(constructor.getParameterTypes(), argDescriptors,
          args)) {
        return constructor;
      }
    }
    return null;
  }

  public static boolean isArgumentMatch(final Class< ? >[] parameterTypes,
      final List<ArgDescriptor> argDescriptors, final Object[] args) {
    int nbrOfArgs = 0;
    if (args != null) {
      nbrOfArgs = args.length;
    }

    int nbrOfParameterTypes = 0;
    if (parameterTypes != null) {
      nbrOfParameterTypes = parameterTypes.length;
    }

    if ((nbrOfArgs == 0) && (nbrOfParameterTypes == 0)) {
      return true;
    }

    if (nbrOfArgs != nbrOfParameterTypes) {
      return false;
    }

    for (int i = 0; i < parameterTypes.length; i++) {
      final Class< ? > parameterType = parameterTypes[i];
      final String argTypeName = argDescriptors != null ? argDescriptors.get(i).getTypeName() : null;
      if (argTypeName != null) {
        if (!argTypeName.equals(parameterType.getName())) {
          return false;
        }
      } else if ((args[i] != null)
          && (!parameterType.isAssignableFrom(args[i].getClass()))) {
        return false;
      }
    }
    return true;
  }

  public static String getSignature(final String methodName,
      final List<ArgDescriptor> argDescriptors, final Object[] args) {
    final StringBuffer buf = new StringBuffer();
    buf.append(methodName).append("(");
    if (args != null) {
      for (int i = 0; i < args.length; i++) {
        String argType = null;
        if (argDescriptors != null) {
          final ArgDescriptor argDescriptor = argDescriptors.get(i);
          if ((argDescriptor != null) && (argDescriptor.getTypeName() != null)) {
            argType = argDescriptor.getTypeName();
          }
        }
        if ((argType == null) && (args[i] != null)) {
          argType = args[i].getClass().getName();
        }
        buf.append(argType);
        if (i < (args.length - 1)) {
          buf.append(", ");
        }
      }
    }
    buf.append(")");
    return buf.toString();
  }

  public static String getUnqualifiedClassName(final Class< ? > clazz) {
    if (clazz == null) {
      return null;
    }
    return ReflectUtil.getUnqualifiedClassName(clazz.getSimpleName());
  }

  public static String getUnqualifiedClassName(String className) {
    if (className == null) {
      return null;
    }
    final int dotIndex = className.lastIndexOf('.');
    if (dotIndex != -1) {
      className = className.substring(dotIndex + 1);
    }
    return className;
  }
}
